import uuid

from fastapi import APIRouter, Depends, HTTPException, Request, Response, status

from wwricu.domain.common import CommonConstant, HttpErrorDetail
from wwricu.domain.input import LoginRO
from wwricu.service.cache import cache_delete, cache_set
from wwricu.service.common import admin_only, hmac_sign, validate_cookie, admin_login


common_api = APIRouter(tags=['Common API'])


@common_api.post('/login')
async def login(login_request: LoginRO, response: Response):
    if admin_login(login_request.username, login_request.password) is not True:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=HttpErrorDetail.WRONG_PASSWORD)
    session_id = uuid.uuid4().hex
    cookie_sign = hmac_sign(session_id)
    await cache_set(session_id, CommonConstant.COOKIE_TIMEOUT_SECOND)
    response.set_cookie(CommonConstant.SESSION_ID, session_id, secure=True, httponly=True, samesite='lax')
    response.set_cookie(CommonConstant.COOKIE_SIGN, cookie_sign, secure=True, httponly=True, samesite='lax')


@common_api.get('/logout', dependencies=[Depends(admin_only)])
async def logout(request: Request, response: Response):
    if (session_id := request.cookies.get(CommonConstant.SESSION_ID)) is None:
        return
    await cache_delete(session_id)
    response.delete_cookie(CommonConstant.SESSION_ID)


@common_api.get('/info', response_model=bool)
async def info(request: Request) -> bool:
    session_id = request.cookies.get(CommonConstant.SESSION_ID)
    cookie_sign = request.cookies.get(CommonConstant.COOKIE_SIGN)
    if valid := await validate_cookie(session_id, cookie_sign):
        await cache_set(session_id, CommonConstant.COOKIE_TIMEOUT_SECOND)
    return valid



import boto3

from wwricu.config import StorageConfig
from wwricu.domain.common import AmazonS3Response


def get_object(key: str) -> bytes | None:
    # If S3 object_name was not specified, use file_name
    result_dict: dict = s3_client.get_object(Bucket=StorageConfig.bucket, Key=key)
    response = AmazonS3Response.model_validate(result_dict)
    return response.Body.read()


def put_object(key: str, data: bytes) -> str:
    s3_client.put_object(Bucket=StorageConfig.bucket, Key=key, Body=data)
    return f'https://{StorageConfig.s3}.{StorageConfig.region}.amazonaws.com/{StorageConfig.bucket}/{key}'


def delete_object(key: str):
    s3_client.delete_object(Bucket=StorageConfig.bucket, Key=key)


s3_client = boto3.client(
    StorageConfig.s3,
    aws_access_key_id= StorageConfig.access_key,
    aws_secret_access_key=StorageConfig.secret_key,
    region_name=StorageConfig.region
)


from typing import Sequence

from sqlalchemy import select, update

from wwricu.domain.entity import BlogPost, PostTag, EntityRelation
from wwricu.domain.enum import TagTypeEnum, RelationTypeEnum
from wwricu.service.database import session


async def get_tags_by_ids(tag_id_list: list[int] = ()) -> list[PostTag]:
    stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_TAG).where(
        PostTag.deleted == False).where(
        PostTag.id.in_(tag_id_list)
    )
    return list((await session.scalars(stmt)).all())


async def get_category_by_id(category_id: int) -> PostTag:
    stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_CAT).where(
        PostTag.deleted == False).where(
        PostTag.id == category_id
    )
    return await session.scalar(stmt)


async def get_category_by_name(category_name: str) -> PostTag | None:
    if category_name is None:
        return None
    stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_CAT).where(
        PostTag.deleted == False).where(
        PostTag.name == category_name
    )
    return await session.scalar(stmt)


async def get_post_ids_by_tag_names(tag_name: list[str]) -> list[int]:
    if not tag_name:
        return []
    stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_TAG).where(
        PostTag.deleted == False).where(
        PostTag.id.in_(tag_name)
    )
    tag_list = (await session.scalar(stmt)).all()
    if not tag_list:
        return []
    stmt = select(EntityRelation).where(
        EntityRelation.type == RelationTypeEnum.POST_TAG).where(
        EntityRelation.deleted == False).where(
        EntityRelation.src_id.in_(tag.id for tag in tag_list)
    )
    return [relation.src_id for relation in (await session.scalar(stmt)).all()]


async def update_tags(post: BlogPost, tag_id_list: list[int] | None = None) -> list[PostTag]:
    if tag_id_list is None:
        return []
    tags = await get_tags_by_ids(tag_id_list)
    stmt = update(EntityRelation).where(
        EntityRelation.type == RelationTypeEnum.POST_TAG).where(
        EntityRelation.deleted == False).where(
        EntityRelation.src_id == post.id
    ).values(deleted=True)
    await session.execute(stmt)
    relations = [EntityRelation(src_id=post.id, dst_id=t.id, type=RelationTypeEnum.POST_TAG) for t in tags]
    session.add_all(relations)
    return tags


async def update_category(post: BlogPost, category_id: int | None = None) -> PostTag | None:
    if category_id is None:
        return None
    category = await get_category_by_id(category_id)
    if category is None:
        return None
    stmt = update(BlogPost).where(
        BlogPost.id == post.id).where(
        BlogPost.deleted == False
    ).values(category_id=category.id)
    await session.execute(stmt)
    return category


async def get_post_tags(post: BlogPost) -> list[PostTag]:
    stmt = select(EntityRelation).where(
        EntityRelation.type == RelationTypeEnum.POST_TAG).where(
        EntityRelation.deleted == False).where(
        EntityRelation.src_id == post.id
    )
    tag_relations = (await session.scalars(stmt)).all()
    return await get_tags_by_ids([relation.dst_id for relation in tag_relations])


async def get_post_category(post: BlogPost) -> PostTag:
    stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_CAT).where(
        PostTag.deleted == False).where(
        PostTag.id == post.category_id
    )
    return await session.scalar(stmt)


async def get_posts_tag_lists(post_list: Sequence[BlogPost]) -> dict[int, list[PostTag]]:
    if not post_list:
        return dict()
    post_tags_dict = {post.id: [] for post in post_list}
    relation_stmt = select(EntityRelation).where(
        EntityRelation.type == RelationTypeEnum.POST_TAG).where(
        EntityRelation.deleted == False).where(
        EntityRelation.src_id.in_(post_tags_dict.keys())
    )
    tag_relations = (await session.scalars(relation_stmt)).all()
    if not tag_relations:
        return dict()
    tag_dict = {tag.id: tag for tag in await get_tags_by_ids([relation.dst_id for relation in tag_relations])}
    for relation in tag_relations:
        post_tags = post_tags_dict.get(relation.src_id)
        tag = tag_dict.get(relation.dst_id)
        post_tags.append(tag)
    return post_tags_dict


async def get_posts_category(post_list: Sequence[BlogPost]) -> dict[int, PostTag]:
    if not post_list:
        return dict()
    cat_stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_CAT).where(
        PostTag.deleted == False).where(
        PostTag.id.in_([post.category_id for post in post_list])
    )
    category_dict = {cat.id: cat for cat in (await session.scalars(cat_stmt)).all()}
    return {post.id: category_dict.get(post.category_id) for post in post_list}



import secrets
import uuid

from wwricu.service.storage import put_object, get_object


def test_storage():
    file = secrets.token_bytes(64)
    key = f'posts/{0}/{uuid.uuid4().hex}'
    put_object(key, file)
    f = get_object(key)
    assert f == file



from dataclasses import field

from wwricu.domain.common import BaseModel
from wwricu.domain.enum import PostStatusEnum, TagTypeEnum


class PostRequestRO(BaseModel):
    page_index: int = 1
    page_size: int = 10
    tag_list: list[str] | None = None
    category: str | None = None
    status: PostStatusEnum | None = None
    deleted: bool = None


class PostUpdateRO(BaseModel):
    """All not null fields are mandatory"""
    id: int
    title: str
    cover_id: int | None = None
    preview: str
    status: PostStatusEnum
    content: str
    tag_id_list: list[int]
    category_id: int | None = None


class PostResourceRO(BaseModel):
    type: str
    payload: str
    name: str | None = None


class LoginRO(BaseModel):
    username: str
    password: str
    captcha: str | None = None
    otp: str | None = None


class TagBatchRO(BaseModel):
    id_list: list[int] = field(default_factory=list)
    type: TagTypeEnum


class TagRO(BaseModel):
    id: int | None = None
    name: str
    type: TagTypeEnum | None = TagTypeEnum.POST_TAG


class TagRequestRO(BaseModel):
    page_size: int | None = 0
    page_index: int | None = 0
    type: TagTypeEnum | None = TagTypeEnum.POST_TAG



from fastapi import APIRouter, Depends

from wwricu.api.common import common_api
from wwricu.api.open import open_api
from wwricu.api.post import post_api
from wwricu.api.tag import tag_api
from wwricu.service.database import database_session


api_router = APIRouter(dependencies=[Depends(database_session)])
api_router.include_router(common_api)
api_router.include_router(open_api)
api_router.include_router(post_api)
api_router.include_router(tag_api)



from enum import StrEnum


class TagTypeEnum(StrEnum):
    POST_TAG: str = 'post_tag'
    POST_CAT: str = 'post_category'


class PostStatusEnum(StrEnum):
    PUBLISHED = 'published'
    DRAFT = 'draft'


class PostResourceTypeEnum(StrEnum):
    IMAGE: str = 'image'
    COVER_IMAGE: str = 'cover'
    COMMON: str = 'common'


class RelationTypeEnum(StrEnum):
    POST_TAG: str = 'post_tag'
    POST_RES: str = 'post_resource'



from sqlalchemy import create_engine, Engine
from wwricu.domain.entity import Base


def test_create_database():
    sync_engine: Engine = create_engine('sqlite:///wwr.sqlite3', echo=True)
    Base.metadata.create_all(sync_engine)



import base64
import hashlib
import hmac
import time
from contextlib import asynccontextmanager
from contextvars import ContextVar

import bcrypt
from fastapi import HTTPException, status, FastAPI
from loguru import logger as log

from wwricu.domain.common import HttpErrorDetail, CommonConstant
from wwricu.config import AdminConfig, Config
from wwricu.service.cache import cache_get


@asynccontextmanager
async def lifespan(_: FastAPI):
    log.info(f'{CommonConstant.APP_TITLE} {CommonConstant.APP_VERSION} Startup')
    log.info(f'listening on {Config.host}:{Config.port}')
    yield
    log.info('THE END')
    await log.complete()


async def admin_login(username: str, password: str) -> bool:
    if __debug__:
        return True
    if username != AdminConfig.username:
        return False
    return bcrypt.checkpw(password.encode(), base64.b64decode(AdminConfig.password))


async def admin_only():
    if admin.get() is not True:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=HttpErrorDetail.NOT_AUTHORIZED)


def hmac_sign(plain: str):
    return hmac.new(secure_key, plain.encode(Config.encoding), hashlib.sha256).hexdigest()


def hmac_verify(plain: str, sign: str) -> bool:
    if not plain or not sign:
        return False
    return hmac_sign(plain) == sign


async def validate_cookie(session_id: str, cookie_sign: str) -> bool:
    if __debug__ is True:
        return True
    if session_id is None or cookie_sign is None or not isinstance(issue_time := await cache_get(session_id), int):
        return False
    if 0 < int(time.time()) - issue_time < CommonConstant.EXPIRE_TIME and hmac_verify(session_id, cookie_sign) is True:
        return True
    return False


admin: ContextVar[bool] = ContextVar('admin', default=False)
secure_key = base64.b64decode(AdminConfig.secure_key.encode(Config.encoding))



from datetime import datetime

from sqlalchemy import String, Integer, Boolean, TEXT, DateTime, func, UniqueConstraint
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from wwricu.domain.common import EntityConstant
from wwricu.domain.enum import RelationTypeEnum, PostStatusEnum


class Base(DeclarativeBase):
    id: Mapped[int] = mapped_column(primary_key=True)
    deleted: Mapped[bool] = mapped_column(Boolean, default=False, index=True)
    create_time: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    update_time: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), onupdate=datetime.now())

    def to_dict(self) -> dict:
        return {key: getattr(self, key) for key in self.__table__.columns.keys()}


class BlogPost(Base):
    __tablename__ = 'wwr_blog_post'
    title: Mapped[str] = mapped_column(String(EntityConstant.USER_STRING_LEN), default='', index=True)
    cover_id: Mapped[int] = mapped_column(Integer(), nullable=True)
    content: Mapped[str] = mapped_column(TEXT(), default='')
    preview: Mapped[str] = mapped_column(TEXT(), default='')
    status: Mapped[str] = mapped_column(
        String(EntityConstant.ENUM_STRING_LEN),
        default=PostStatusEnum.DRAFT.value,
        index=True
    )
    category_id: Mapped[int] = mapped_column(Integer(), nullable=True)


class PostTag(Base):
    __tablename__ = 'wwr_post_tag'
    __table_args__ = (UniqueConstraint('name', 'type', name='uix_name_type'),)
    name: Mapped[str] = mapped_column(String(EntityConstant.USER_STRING_LEN))
    type: Mapped[str] = mapped_column(String(EntityConstant.ENUM_STRING_LEN))


class EntityRelation(Base):
    __tablename__ = 'wwr_entity_tag_relation'
    src_id: Mapped[int] = mapped_column(Integer(), index=True)
    dst_id: Mapped[int] = mapped_column(Integer(), index=True)
    type: Mapped[RelationTypeEnum] = mapped_column(
        String(EntityConstant.ENUM_STRING_LEN),
        default=RelationTypeEnum.POST_TAG
    )


class PostResource(Base):
    __tablename__ = 'wwr_post_resource'
    post_id: Mapped[int] = mapped_column(Integer(), index=True)
    name: Mapped[str] = mapped_column(
        String(EntityConstant.USER_STRING_LEN),
        nullable=True,
        comment='File original name'
    )
    key: Mapped[str] = mapped_column(String(EntityConstant.LONG_STRING_LEN), comment='OSS Key')
    type: Mapped[str] = mapped_column(String(EntityConstant.ENUM_STRING_LEN), nullable=True)
    url: Mapped[str] = mapped_column(TEXT(), nullable=True, comment='Public url')



from loguru import logger as log
from fastapi import status

from sqlalchemy import select

from test.test_client import client
from wwricu.domain.entity import PostTag, EntityRelation, BlogPost
from wwricu.domain.enum import TagTypeEnum
from wwricu.service.post import get_all_post_details


def test_create_post():
    response = client.get('/post/create')
    assert response.status_code == status.HTTP_200_OK
    log.info(response.json())


def test_get_post_detail():
    response = client.get('/post/detail/1')
    assert response.status_code == status.HTTP_200_OK
    log.info(response.json())


def test_get_all_post():
    get_all_post_details([])
    cat_stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_CAT).where(
        PostTag.deleted == False).where(
        PostTag.id.in_((1, 2))
    )
    tag_stmt = select(PostTag).where(
        PostTag.type == TagTypeEnum.POST_TAG).where(
        PostTag.deleted == False).join(
        EntityRelation, PostTag.id == EntityRelation.dst_id).join(
        BlogPost, EntityRelation.src_id == BlogPost.id
    )
    print(tag_stmt)
    print(cat_stmt)



from datetime import datetime

from botocore.response import StreamingBody
from loguru import logger as log
from pydantic import BaseModel as PydanticBaseModel, ConfigDict


class BaseModel(PydanticBaseModel):
    model_config =  ConfigDict(from_attributes=True)


class HttpErrorDetail(object):
    POST_NOT_FOUND: str = 'Post Not Found'
    NO_SUCH_USER: str = 'NO SUCH USER'
    WRONG_PASSWORD: str = 'WRONG PASSWORD'
    NOT_AUTHORIZED: str = 'Not authorized'
    UPLOAD_FAILURE: str = 'Failed to upload'
    INVALID_TAG_TYPE: str = 'Invalid tag type'


class CommonConstant(object):
    SESSION_ID: str = 'session_id'
    COOKIE_SIGN: str = 'sign'
    EXPIRE_TIME: int = 60 * 60 * 24 * 7
    APP_TITLE: str = 'wwr.icu'
    APP_VERSION: str = 'v2.0.0'
    CONFIG_PATH: str = 'conf/config.json'
    TOKEN_PATH: str = 'conf/github_token.txt'
    STORE_RET_KEY: str = 'key'
    COOKIE_TIMEOUT_SECOND: int = 30 * 24 * 60 * 60


class EntityConstant(object):
    ENUM_STRING_LEN: int = 32
    USER_STRING_LEN: int = 64
    LONG_STRING_LEN: int = 128


class ConfigCenterConst(object):
    URL: str = 'https://api.github.com/repos/wwricu/config/contents/config.json'
    ACCEPT: str = 'application/vnd.github+json'
    AUTHORIZATION: str = 'Bearer {token}'
    TOKEN_KEY: str = 'github_token'


class AmazonS3ResponseMetaData(BaseModel):
    RequestId: str
    HostId: str
    HTTPStatusCode: int
    HTTPHeaders: dict


class AmazonS3Response(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)
    ResponseMetadata: dict
    AcceptRanges: str
    LastModified: datetime
    ContentLength: int
    ETag: str
    ContentType: str
    ServerSideEncryption: str
    Metadata: dict
    ResponseMetadata: dict
    Body: StreamingBody


class GithubContentResponse(BaseModel):
    name: str
    content: str
    download_url: str


def retry(max_try: int = 3):
    def decorator(func: callable):
        def wrapper(*args, **kwargs):
            t = max_try
            while True:
                try:
                    t -= 1
                    return func(*args, **kwargs)
                except Exception as e:
                    if t == 0:
                        raise e
                    log.info(f'{func.__name__} retry {max_try - t} times')
        return wrapper
    return decorator



import asyncio
import uuid

from fastapi import APIRouter, Depends, Form, HTTPException, status as http_status, UploadFile

from sqlalchemy import desc, func, select, update

from wwricu.domain.common import HttpErrorDetail
from wwricu.domain.entity import BlogPost, EntityRelation, PostResource
from wwricu.domain.enum import PostStatusEnum, PostResourceTypeEnum
from wwricu.domain.input import PostUpdateRO, PostRequestRO
from wwricu.domain.output import PostDetailVO, FileUploadVO, PostDetailPageVO
from wwricu.service.common import admin_only
from wwricu.service.database import session
from wwricu.service.post import get_post_by_id, delete_post_cover, get_all_post_details, get_post_detail
from wwricu.service.storage import put_object
from wwricu.service.tag import update_category, update_tags, get_category_by_name, get_post_ids_by_tag_names


post_api = APIRouter(prefix='/post', tags=['Post Management'], dependencies=[Depends(admin_only)])


@post_api.get('/create', response_model=PostDetailVO)
async def create_post() -> PostDetailVO:
    blog_post = BlogPost(status=PostStatusEnum.DRAFT)
    session.add(blog_post)
    await session.flush()
    return PostDetailVO.model_validate(blog_post)


@post_api.post('/all', response_model=PostDetailPageVO)
async def select_post(post: PostRequestRO) -> PostDetailPageVO:
    stmt = select(BlogPost)
    if post.status is not None:
        stmt = stmt.where(BlogPost.status == post.status.value)
    if post.deleted is not None:
        stmt = stmt.where(BlogPost.deleted == post.deleted)
    if category := await get_category_by_name(post.category):
        stmt = stmt.where(BlogPost.category_id == category.id)
    if post.tag_list is not None:
        post_id_list = await get_post_ids_by_tag_names(post.tag_list)
        stmt = stmt.where(BlogPost.id.in_(post_id_list))
    count_stmt = select(func.count()).select_from(stmt.subquery())
    post_stmt = stmt.order_by(
        desc(BlogPost.create_time)).limit(
        post.page_size).offset(
        (post.page_index - 1) * post.page_size
    )

    posts_result, count = await asyncio.gather(session.scalars(post_stmt), session.scalar(count_stmt))
    all_posts = await get_all_post_details(posts_result.all())
    return PostDetailPageVO(page_index=post.page_index, page_size=post.page_size, count=count, post_details=all_posts)


@post_api.get('/detail/{post_id}', response_model=PostDetailVO | None)
async def get_post(post_id: int) -> PostDetailVO | None:
    if (post := await get_post_by_id(post_id)) is None:
        return None
    return await get_post_detail(post)


@post_api.post('/update', response_model=PostDetailVO)
async def update_post(post_update: PostUpdateRO) -> PostDetailVO:
    if (blog_post := await get_post_by_id(post_update.id)) is None:
        raise HTTPException(status_code=http_status.HTTP_404_NOT_FOUND, detail=HttpErrorDetail.POST_NOT_FOUND)
    if blog_post.cover_id is not None and blog_post.cover_id != post_update.cover_id:
        await delete_post_cover(blog_post)
    stmt = update(BlogPost).where(BlogPost.id == post_update.id).values(
        title=post_update.title,
        content=post_update.content,
        preview=post_update.preview,
        cover_id=post_update.cover_id,
        status=post_update.status,
        category_id=post_update.category_id
    )
    await session.execute(stmt)
    await update_category(blog_post, post_update.category_id)
    await update_tags(blog_post, post_update.tag_id_list)
    return await get_post_detail(blog_post)


@post_api.get('/status/{post_id}', response_model=PostDetailVO)
async def update_post_status(post_id: int, status: str) -> PostDetailVO:
    if (blog_post := await get_post_by_id(post_id)) is None:
        raise HTTPException(status_code=http_status.HTTP_404_NOT_FOUND, detail=HttpErrorDetail.POST_NOT_FOUND)
    stmt = update(BlogPost).where(BlogPost.id == blog_post.id).values(status=PostStatusEnum(status))
    await session.execute(stmt)
    return await get_post_detail(blog_post)


@post_api.get('/delete/{post_id}', response_model=int)
async def delete_post(post_id: int):
    stmt = update(BlogPost).where(BlogPost.id == post_id).values(deleted=True)
    tag_stmt = update(EntityRelation).where(EntityRelation.src_id == post_id).values(deleted=True)
    result = await session.execute(stmt)
    await session.execute(tag_stmt)
    return result.rowcount


@post_api.post('/upload', response_model=FileUploadVO)
async def upload(file: UploadFile, post_id: int = Form(), file_type: str = Form()) -> FileUploadVO:
    if (post := await get_post_by_id(post_id)) is None:
        raise HTTPException(status_code=http_status.HTTP_404_NOT_FOUND, detail=HttpErrorDetail.POST_NOT_FOUND)
    type_enum = PostResourceTypeEnum(file_type)
    key = f'post/{post_id}/{type_enum}_{uuid.uuid4().hex}'
    url = put_object(key, await file.read())
    resource = PostResource(name=file.filename, key=key, type=type_enum, post_id=post.id, url=url)
    session.add(resource)
    await session.flush()
    return FileUploadVO(id=resource.id, name=file.filename, key=key, location=url)



import time
from typing import override

from loguru import logger as log
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request

from wwricu.service.common import validate_cookie, admin
from wwricu.domain.common import CommonConstant


class PerformanceMiddleware(BaseHTTPMiddleware):
    @override
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        begin = time.time()
        try:
            return await call_next(request)
        finally:
            log.info(f'{request.method} {request.url.path} {(int((time.time() - begin) * 1000))} ms')


class AspectMiddleware(BaseHTTPMiddleware):
    @override
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        log.info(f'{request.method} {request.url.path}')
        try:
            return await call_next(request)
        except Exception as e:
            log.exception(e)


class AuthMiddleware(BaseHTTPMiddleware):
    @override
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        session_id = request.cookies.get(CommonConstant.SESSION_ID)
        cookie_sign = request.cookies.get(CommonConstant.COOKIE_SIGN)
        if await validate_cookie(session_id, cookie_sign):
            admin.set(True)
        try:
            return await call_next(request)
        finally:
            admin.set(False)


# noinspection PyTypeChecker
middlewares = [
    Middleware(AspectMiddleware),
    Middleware(AuthMiddleware)
]


if __debug__ is True:
    # noinspection PyTypeChecker
    middlewares.insert(0, Middleware(PerformanceMiddleware))
    # noinspection PyTypeChecker
    middlewares.append(Middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
        expose_headers=['*']
    ))



import asyncio

from fastapi import APIRouter, HTTPException, status
from sqlalchemy import select, desc, func

from wwricu.domain.common import HttpErrorDetail
from wwricu.domain.entity import BlogPost, PostTag
from wwricu.domain.enum import PostStatusEnum
from wwricu.domain.input import PostRequestRO, TagRequestRO
from wwricu.domain.output import TagVO, PostDetailVO, PostDetailPageVO
from wwricu.service.database import session
from wwricu.service.post import get_all_post_details
from wwricu.service.tag import get_category_by_name, get_post_ids_by_tag_names

open_api = APIRouter(prefix='/open', tags=['Open API'])


@open_api.post('/post/all', response_model=PostDetailPageVO)
async def get_all_posts(post: PostRequestRO) -> PostDetailPageVO:
    stmt = select(
        BlogPost.id,
        BlogPost.title,
        BlogPost.preview,
        BlogPost.cover_id,
        BlogPost.category_id,
        BlogPost.create_time,
        BlogPost.update_time
    ).where(
        BlogPost.deleted == False).where(
        BlogPost.status == PostStatusEnum.PUBLISHED
    )
    if category := await get_category_by_name(post.category):
        stmt = stmt.where(BlogPost.category_id == category.id)
    if post.tag_list is not None:
        post_id_list = await get_post_ids_by_tag_names(post.tag_list)
        stmt = stmt.where(BlogPost.id.in_(post_id_list))
    count_stmt = select(func.count()).select_from(stmt.subquery())
    post_stmt = stmt.order_by(
        desc(BlogPost.create_time)).limit(
        post.page_size).offset(
        (post.page_index - 1) * post.page_size
    )
    posts_result, count = await asyncio.gather(session.execute(post_stmt), session.scalar(count_stmt))
    all_posts = await get_all_post_details(posts_result.all())
    return PostDetailPageVO(page_index=post.page_index, page_size=post.page_size, count=count, post_details=all_posts)


@open_api.get('/post/detail/{post_id}', response_model=PostDetailVO)
async def get_post_detail(post_id: int) -> PostDetailVO:
    stmt = select(
        BlogPost.id,
        BlogPost.content,
        BlogPost.cover_id,
        BlogPost.category_id,
        BlogPost.create_time,
        BlogPost.update_time
    ).where(
        BlogPost.id == post_id).where(
        BlogPost.deleted == False).where(
        BlogPost.status == PostStatusEnum.PUBLISHED
    )
    if (post := await session.scalar(stmt)) is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=HttpErrorDetail.POST_NOT_FOUND)
    return await get_post_detail(post)


@open_api.post('/tags')
async def get_all_tags(get_tag: TagRequestRO) -> list[TagVO]:
    stmt = select(PostTag).where(
        PostTag.deleted == False).where(
        PostTag.type == get_tag.type).order_by(
        desc(PostTag.update_time)
    )
    if get_tag.page_index > 0 and get_tag.page_size > 0:
        stmt = stmt.limit(get_tag.page_size).offset((get_tag.page_index - 1) * get_tag.page_size)
    return [TagVO.model_validate(tag) for tag in (await session.scalars(stmt)).all()]



from fastapi import APIRouter, Depends, HTTPException, status
from sqlalchemy import update, select

from wwricu.domain.common import HttpErrorDetail
from wwricu.domain.entity import PostTag, EntityRelation, BlogPost
from wwricu.domain.enum import TagTypeEnum, RelationTypeEnum
from wwricu.domain.input import TagRO, TagBatchRO
from wwricu.domain.output import TagVO
from wwricu.service.common import admin_only
from wwricu.service.database import session


tag_api = APIRouter(prefix='/tag', tags=['Tag Management'], dependencies=[Depends(admin_only)])


@tag_api.post('/create', response_model=TagVO)
async def create_tag(tag_create: TagRO) -> TagVO:
    tag = PostTag(
        name=tag_create.name,
        type=tag_create.type
    )
    session.add(tag)
    await session.flush()
    return TagVO.model_validate(tag)


@tag_api.post('/update', response_model=TagVO)
async def update_tag(tag_update: TagRO) -> TagVO:
    stmt = update(PostTag).where(PostTag.id == tag_update.id).values(name=tag_update.name)
    await session.execute(stmt)
    stmt = select(PostTag).where(PostTag.id == tag_update.id)
    return TagVO.model_validate(await session.scalar(stmt))


@tag_api.post('/delete', response_model=int)
async def delete_tags(tag_batch: TagBatchRO):
    post_stmt = update(PostTag).where(
        PostTag.type == tag_batch.type).where(
        PostTag.id.in_(tag_batch.id_list)).values(
        deleted=True
    )
    if tag_batch.type == TagTypeEnum.POST_TAG:
        tag_stmt = update(EntityRelation).where(
            EntityRelation.type == RelationTypeEnum.POST_TAG).where(
            EntityRelation.dst_id.in_(tag_batch.id_list)
        ).values(deleted=True)
    elif tag_batch.type == TagTypeEnum.POST_CAT:
        tag_stmt = update(BlogPost).where(BlogPost.category_id.in_(tag_batch.id_list)).values(category_id=None)
    else:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=HttpErrorDetail.INVALID_TAG_TYPE)
    result = await session.execute(post_stmt)
    await session.execute(tag_stmt)
    return result.rowcount



from __future__ import annotations

from dataclasses import field
from datetime import datetime

from wwricu.domain.common import BaseModel
from wwricu.domain.entity import PostTag
from wwricu.domain.enum import TagTypeEnum, PostResourceTypeEnum, PostStatusEnum


class TagVO(BaseModel):
    id: int
    name: str
    type: TagTypeEnum

    @classmethod
    def of(cls, tag: PostTag) -> TagVO | None:
        if tag is None:
            return None
        return cls(id=tag.id, name=tag.name, type=TagTypeEnum(tag.type))


class PostResourceVO(BaseModel):
    id: int
    name: str | None = None
    key: str
    url: str
    type: PostResourceTypeEnum


class PostDetailVO(BaseModel):
    id: int
    title: str | None = None
    cover: PostResourceVO | None = None
    preview: str = ''
    content: str = ''
    tag_list: list[TagVO] = field(default_factory=list)
    category: TagVO | None = None
    status: PostStatusEnum | None = None
    create_time: datetime | None = None
    update_time: datetime | None = None


class PostDetailPageVO(BaseModel):
    page_index: int
    page_size: int
    count: int
    post_details: list[PostDetailVO] = field(default_factory=list)


class FileUploadVO(BaseModel):
    id: int
    name: str
    key: str
    location: str



import uvicorn
from fastapi import FastAPI

from wwricu.api import api_router
from wwricu.config import Config
from wwricu.domain.common import CommonConstant
from wwricu.service.common import lifespan
from wwricu.middleware import middlewares


app = FastAPI(
    title=CommonConstant.APP_TITLE,
    version=Config.version,
    lifespan=lifespan,
    middleware=middlewares,
    debug=__debug__
)
app.include_router(api_router)


def main():
    uvicorn.run(app=Config.app, host=Config.host, port=Config.port, log_level=Config.log_level)



import base64
import json
import os
from logging import CRITICAL

import httpx
from httpx import Response
from loguru import logger as log

from wwricu.domain.common import CommonConstant, ConfigCenterConst, GithubContentResponse, retry


class ConfigClass(object):
    @classmethod
    def init(cls, **kwargs):
        for k, v in kwargs.items():
            if k in cls.__annotations__:
                setattr(cls, k, v)


class StorageConfig(ConfigClass):
    region: str
    s3: str = 's3'
    access_key: str
    secret_key: str
    bucket: str


class DatabaseConfig(ConfigClass):
    engine: str
    username: str
    password: str
    host: str
    port: int
    database: str
    url: str

    @classmethod
    def get_url(cls):
        try:
            return cls.url
        except (Exception,):
            return f'{cls.engine}://{cls.username}:{cls.password}@{cls.host}:{cls.port}/{cls.database}'


class AdminConfig(ConfigClass):
    username: str
    password: str
    secure_key: str
    secure_key_bytes: bytes


class Config(ConfigClass):
    app: str = 'main:app'
    host: str = '0.0.0.0'
    port: int = 8000
    log_level: int = CRITICAL
    encoding: str = 'utf-8'
    version_file: str = 'version.txt'
    version: str = 'stable'

    @classmethod
    def load(cls, admin_config: dict, database_config: dict, storage_config: dict, **kwargs):
        cls.init(**kwargs)
        with open(cls.version_file) as f:
            cls.version = f.readline().strip()
        AdminConfig.init(**admin_config)
        DatabaseConfig.init(**database_config)
        StorageConfig.init(**storage_config)


@retry()
def download_config():
    try:
        with open(CommonConstant.TOKEN_PATH) as f:
            token = f.readline()
    except (Exception,):
        token = os.environ.get(ConfigCenterConst.TOKEN_KEY)
    if not token:
        log.info('Config center disabled')
        return
    response: Response = httpx.get(ConfigCenterConst.URL, headers=dict(
        Accept=ConfigCenterConst.ACCEPT,
        Authorization=ConfigCenterConst.AUTHORIZATION.format(token=token.strip())
    ))
    github_response = GithubContentResponse.model_validate(response.json())
    content: bytes = base64.b64decode(github_response.content.encode())
    with open(CommonConstant.CONFIG_PATH, 'wb+') as f:
        f.write(content)
        log.info(f'Download {ConfigCenterConst.URL} to {CommonConstant.CONFIG_PATH}')


def init():
    if __debug__:
        log.warning('APP RUNNING ON DEBUG MODE')
    try:
        download_config()
    except Exception as e:
        log.warning(f'Failed to download config: {e}')
    with open(CommonConstant.CONFIG_PATH) as conf:
        Config.load(**json.load(conf))
        log.info('Config init')



import asyncio
from typing import Sequence

from sqlalchemy import delete, select

from wwricu.domain.output import PostDetailVO, TagVO, PostResourceVO
from wwricu.service.storage import delete_object
from wwricu.domain.entity import BlogPost, PostResource
from wwricu.domain.enum import PostResourceTypeEnum
from wwricu.service.database import session
from wwricu.service.tag import get_post_category, get_post_tags, get_posts_category, get_posts_tag_lists


async def get_post_by_id(post_id: int) -> BlogPost:
    stmt = select(BlogPost).where(BlogPost.id == post_id).where(BlogPost.deleted == False)
    return await session.scalar(stmt)


async def get_post_cover(post: BlogPost) -> PostResource:
    stmt = select(PostResource).where(
        PostResource.deleted == False).where(
        PostResource.type == PostResourceTypeEnum.COVER_IMAGE).where(
        PostResource.id == post.cover_id
    )
    return await session.scalar(stmt)


async def delete_post_cover(post: BlogPost) -> int:
    """HARD DELETE the resource because we are using free object storage"""
    stmt = select(PostResource).where(
        PostResource.deleted == False).where(
        PostResource.type == PostResourceTypeEnum.COVER_IMAGE).where(
        PostResource.id == post.cover_id
    )
    if (resource := await session.scalar(stmt)) is None:
        return 0
    delete_object(resource.key)
    stmt = delete(PostResource).where(PostResource.id == resource.id)
    result = await session.execute(stmt)
    return result.rowcount


async def clean_post_resource():
    """delete all unused files from oss"""
    pass


async def get_posts_cover(post_list: Sequence[BlogPost]) -> dict[int, PostResource]:
    cat_stmt = select(PostResource).where(
        PostResource.type == PostResourceTypeEnum.COVER_IMAGE).where(
        PostResource.deleted == False).where(
        PostResource.id.in_(post.cover_id for post in post_list)
    )
    resource_dict = {res.id: res for res in (await session.scalars(cat_stmt)).all()}
    return {post.id: resource_dict.get(post.cover_id) for post in post_list}


async def get_post_detail(blog_post: BlogPost) -> PostDetailVO:
    category, tags, cover = await asyncio.gather(
        get_post_category(blog_post),
        get_post_tags(blog_post),
        get_post_cover(blog_post)
    )
    post_detail = PostDetailVO.model_validate(blog_post)
    post_detail.tag_list = [TagVO.model_validate(tag) for tag in tags]
    if category is not None:
        post_detail.category = TagVO.model_validate(category)
    if cover is not None:
        post_detail.cover = PostResourceVO.model_validate(cover)
    return post_detail


async def get_all_post_details(post_list: Sequence[BlogPost]) -> list[PostDetailVO]:
    categories, tags = await asyncio.gather(
        get_posts_category(post_list),
        get_posts_tag_lists(post_list)
    )
    result = []
    for post in post_list:
        detail = PostDetailVO.model_validate(post)
        if category := categories.get(post.id):
            detail.category = TagVO.model_validate(category)
        if tag_list := tags.get(post.id):
            detail.tag_list = TagVO.model_validate(tag_list)
        result.append(detail)
    return result



import asyncio


async def timeout(key: str, second: int):
    await asyncio.sleep(second)
    await cache_delete(key)


async def cache_delete(key: str):
    cache_data.pop(key)
    if task := timeout_callback.pop(key, None):
        task.cancel()


async def cache_set(key: str, value: any, second: int | None = 600):
    if task := timeout_callback.pop(key, None):
        task.cancel()
    cache_data[key] = value
    if second is not None and second > 0:
        timeout_callback[key] = asyncio.create_task(timeout(key, second))


async def cache_get(key: str) -> any:
    return cache_data.get(key)


cache_data: dict[str, any] = dict()
timeout_callback: dict[str, asyncio.Task] = dict()



from asyncio import current_task
from contextlib import asynccontextmanager

from sqlalchemy.ext.asyncio import async_scoped_session, async_sessionmaker, create_async_engine, AsyncSession

from wwricu.config import DatabaseConfig


async def database_session() -> AsyncSession:
    try:
        yield session
    finally:
        if session.registry.registry:
            await session.commit()
            await session.remove()


@asynccontextmanager
async def new_session() -> AsyncSession:
    async with AsyncSession(engine) as s, s.begin():
        yield s


def sync_create_database():
    from sqlalchemy import create_engine, Engine
    from wwricu.domain.entity import Base
    sync_engine: Engine = create_engine(f'sqlite:///{DatabaseConfig.database}', echo=True)
    Base.metadata.create_all(sync_engine)


# sync_create_database()
engine = create_async_engine(DatabaseConfig.get_url(), echo=__debug__)
session_maker = async_sessionmaker(bind=engine, expire_on_commit=False)
session = async_scoped_session(session_maker, scopefunc=current_task)
